#!/usr/bin/env python

'''Driver UI'''

import tkinter as tk
from tkinter import ttk

from ac_websocket_client.objects import (
    GriddedButton, GriddedEntry, GriddedFrame, GriddedLabel, TrafficLight)


class DriverUI(GriddedFrame):
    '''Driver UI'''

    REORDER_OPTIONS = [
        'n/a',
        "Starting",
        "Finishing",
        "Reversed"
    ]

    def __init__(self, parent):

        super().__init__(grid_row=3, grid_col=0, height_by=4.25)

        self.parent = parent

        self.configure_columns(1, 1, 4, 1, 1)

        self._buttons = {}
        self._entries = {}
        self._lights = {}
        self._slots = ['n/a']

        GriddedLabel(self, grid_row=0, grid_col=0, width=8, text="Drivers")

        self._entry = tk.StringVar()
        GriddedEntry(self, grid_row=0, grid_col=2,
                     textvariable=self._entry, state=tk.DISABLED)

        GriddedLabel(self, grid_row=0, grid_col=1,
                     width=8, text="ini:")

        self._buttons['save'] = tk.StringVar(value='Save .ini')
        GriddedButton(self, grid_row=0, grid_col=3,
                      textvariable=self._buttons['save'],
                      command=lambda: self.parent.loop.create_task(self.parent.update_grid(write=True)))
        self._lights['save'] = TrafficLight(self, row=0, column=4)
        self._lights['save'].gray()

        GriddedLabel(self, grid_row=1, grid_col=1,
                     width=8, text="order by:")

        self._order_by = tk.StringVar()
        self._order_options = ttk.OptionMenu(
            self, self._order_by, *DriverUI.REORDER_OPTIONS)
        self._order_options.grid(row=1, column=2, padx=15, sticky='ew')
        self._order_by.set('n/a')

        GriddedButton(self, grid_row=1, grid_col=3,
                      text='Modify order', width=10,
                      command=lambda: self.parent.loop.create_task(self._upgrade_grid()))

        GriddedLabel(self, grid_row=2, grid_col=1,
                     width=8, text="slot:")

        self._slot_chosen = tk.StringVar()
        self._slot_options = ttk.OptionMenu(
            self, self._slot_chosen, *self._slots, command=self._update_slots)
        self._slot_options.grid(row=2, column=2, padx=15, sticky='ew')
        self._slot_chosen.set('n/a')

        GriddedButton(self, grid_row=2, grid_col=3,
                      text='Modify Slot', width=10,
                      command=lambda: self.parent.loop.create_task(self._not_implemented()))

        GriddedLabel(self, grid_row=3, grid_col=1,
                     width=8, text="car_id:")
        self._entries['car_id'] = tk.StringVar()
        GriddedEntry(self, grid_row=3, grid_col=2,
                     textvariable=self._entries['car_id'])

        GriddedLabel(self, grid_row=4, grid_col=1,
                     width=8, text="model:")
        self._entries['model'] = tk.StringVar()
        GriddedEntry(self, grid_row=4, grid_col=2,
                     textvariable=self._entries['model'])

        GriddedLabel(self, grid_row=5, grid_col=1,
                     width=8, text="guid:")
        self._entries['guid'] = tk.StringVar()
        GriddedEntry(self, grid_row=5, grid_col=2,
                     textvariable=self._entries['guid'])

        GriddedLabel(self, grid_row=6, grid_col=1,
                     width=8, text="drivername:")
        self._entries['drivername'] = tk.StringVar()
        GriddedEntry(self, grid_row=6, grid_col=2,
                     textvariable=self._entries['drivername'])

        GriddedLabel(self, grid_row=7, grid_col=1,
                     width=8, text="team:")
        self._entries['team'] = tk.StringVar()
        GriddedEntry(self, grid_row=7, grid_col=2,
                     textvariable=self._entries['team'])

        GriddedLabel(self, grid_row=8, grid_col=1,
                     width=8, text="skin:")
        self._entries['skin'] = tk.StringVar()
        GriddedEntry(self, grid_row=8, grid_col=2,
                     textvariable=self._entries['skin'])

        GriddedLabel(self, grid_row=9, grid_col=1,
                     width=8, text="ballast:")
        self._entries['ballast'] = tk.StringVar()
        GriddedEntry(self, grid_row=9, grid_col=2,
                     textvariable=self._entries['ballast'])

        GriddedLabel(self, grid_row=10, grid_col=1,
                     width=8, text="restrictor:")
        self._entries['restrictor'] = tk.StringVar()
        GriddedEntry(self, grid_row=10, grid_col=2,
                     textvariable=self._entries['restrictor'])

        self.update_ui()

    async def _upgrade_grid(self):

        match self._order_by.get():
            case "Starting":
                await self.parent.update_grid(by_starting=True)
            case "Finishing":
                await self.parent.update_grid(by_finishing=True)
            case "Reversed":
                await self.parent.update_grid(by_reversed=True)

        self.parent.states.ini_needs_saving = True

        self.update_ui()

    def _update_slots(self, slot):

        if slot != 'n/a':
            if entry := self.parent.entries.get(slot, None):
                self._entries['car_id'].set(entry['car_id'])
                self._entries['model'].set(entry['model'])
                self._entries['drivername'].set(entry['drivername'])
                self._entries['guid'].set(entry['guid'])
                self._entries['team'].set(entry['team'])
                self._entries['skin'].set(entry['skin'])
                self._entries['ballast'].set(entry['ballast'])
                self._entries['restrictor'].set(entry['restrictor'])
        else:
            self._entries['car_id'].set('')
            self._entries['model'].set('')
            self._entries['drivername'].set('')
            self._entries['guid'].set('')
            self._entries['team'].set('')
            self._entries['skin'].set('')
            self._entries['ballast'].set('')
            self._entries['restrictor'].set('')

    def update_ui(self):
        '''Update the UI based on states'''

        if self.parent.states.is_connected:

            self._slots = []
            for key in self.parent.entries:
                self._slots.append(key)

            self._slot_options.set_menu(0, *self._slots)

            if self.parent.states.ini_needs_saving:
                self._lights['save'].red()
            else:
                self._lights['save'].gray()

        else:

            self._order_by.set('n/a')
            self._slot_chosen.set('n/a')
            self._update_slots('n/a')



#!/usr/bin/env python

'''Assetto Corsa Websockets App Objects'''

import asyncio
import tkinter as tk
from tkinter import ttk

APP_TITLE = 'ACWS Client'
BUTTON_X = 10
COLUMN_X = 50
CONSOLE_X = 86
CONSOLE_Y = 15
ENTRY_X = 40
FRAME_X = 800
FRAME_Y = 100
LABEL_X = 8
PAD_X = 10
PAD_Y = 5


class GriddedButton(ttk.Button):
    '''Button in a grid'''

    def __init__(self, parent, grid_row: int, grid_col: int,
                 width: int = BUTTON_X, **kwargs):

        super().__init__(parent, width=width, **kwargs)

        self.grid_propagate(False)
        self.grid(row=grid_row, column=grid_col, padx=PAD_X, pady=PAD_Y)


class GriddedEntry(ttk.Entry):
    '''Entry in a grid'''

    def __init__(self, parent,
                 grid_row: int, grid_col: int,
                 width: int = ENTRY_X, **kwargs):

        super().__init__(parent, width=width, foreground='black', **kwargs)

        self.grid_propagate(False)
        self.grid(row=grid_row, column=grid_col, padx=PAD_X, pady=PAD_Y)


class GriddedFrame(tk.Frame):
    '''Frame inside a Grid'''

    def __init__(self, grid_row: int, grid_col: int,
                 height_by: float = 1, width_by: float = 1, **kwargs):

        super().__init__(height=FRAME_Y * height_by,
                         width=FRAME_X * width_by, **kwargs)

        self.grid_propagate(False)
        self.grid(row=grid_row, column=grid_col, padx=PAD_X, pady=PAD_Y)

    def configure_columns(self, *weights):
        '''Helper function to configure column weights'''
        i = 0
        for weight in weights:
            self.columnconfigure(i, weight=weight)
            i += 1

    def configure_rows(self, *weights):
        '''Helper function to configure row weights'''
        i = 0
        for weight in weights:
            self.rowconfigure(i, weights=weight)
            i += 1


class GriddedLabel(ttk.Label):
    '''Label inside a Grid'''

    def __init__(self, parent, grid_row: int, grid_col: int, width: int = LABEL_X, **kwargs):

        super().__init__(parent, width=width, **kwargs)

        self.grid_propagate(False)
        self.grid(row=grid_row, column=grid_col, padx=PAD_X, pady=PAD_Y)


class GriddedListbox(tk.Listbox):
    '''Listbox in a grid'''

    def __init__(self, parent, grid_row: int, grid_col: int,
                 grid_span: int = 1, **kwargs):

        super().__init__(parent, height=CONSOLE_Y, width=CONSOLE_X, **kwargs)

        scrollbar = tk.Scrollbar(parent, orient='vertical')
        scrollbar.config(command=self.yview)

        self.config(yscrollcommand=scrollbar.set)

        self.grid(row=grid_row, column=grid_col,
                  columnspan=grid_span, padx=PAD_X, pady=PAD_Y)
        scrollbar.grid(row=grid_row, column=grid_span-1, sticky='ns')


class GriddedText(tk.Text):
    '''Text in a gride'''

    def __init__(self, parent, grid_row: int, grid_col: int,
                 grid_span: int = 1, **kwargs):

        super().__init__(parent,
                         height=CONSOLE_Y * 75,
                         width=CONSOLE_X * 75,
                         wrap='word', **kwargs)

        scrollbar = tk.Scrollbar(parent, orient='vertical')
        scrollbar.config(command=self.yview)

        self.config(yscrollcommand=scrollbar.set)

        self.grid(row=grid_row, column=grid_col,
                  columnspan=grid_span, padx=PAD_X, pady=PAD_Y)
        scrollbar.grid(row=grid_row, column=grid_span-1, sticky='ns')


class GriddedTreeview(ttk.Treeview):
    '''Treeview in a grid'''

    def __init__(self, parent, grid_row: int, grid_col: int,
                 grid_span: int = 1, **kwargs):

        super().__init__(parent, height=CONSOLE_Y - 2, show='headings', **kwargs)

        scrollbar = tk.Scrollbar(parent, orient='vertical')
        scrollbar.config(command=self.yview)

        self.config(yscrollcommand=scrollbar.set)

        self.grid(row=grid_row, column=grid_col,
                  columnspan=grid_span, padx=PAD_X, pady=PAD_Y)
        scrollbar.grid(row=grid_row, column=grid_span-1, sticky='ns')

        self._columns = ()

    def add_columns(self, *cols):
        '''Add all columns to the tree'''

        self['columns'] = cols
        for col in cols:
            self.column(col, anchor=tk.W)
            self.heading(col, text=col, anchor=tk.W)

    def set_widths(self, *widths):
        '''Set column widths'''

        i = 0
        for width in widths:
            self.column(i, minwidth=width, width=width)
            i += 1


class TrafficLight():
    '''Tk based traffic light'''

    def __init__(self, parent, row: int, column: int):
        self.colour = 'red'
        self.canvas = tk.Canvas(parent, width=20, height=20)
        self.canvas.grid(row=row, column=column, padx=PAD_X, pady=PAD_Y)
        self.light = self.canvas.create_oval(5, 5, 20, 20, fill=self.colour)
        self._fill = None
        self._flash = False
        self._timer = None

    def gray(self):
        self.canvas.itemconfigure(self.light, fill='darkgray')

    def green(self):
        self.canvas.itemconfigure(self.light, fill='green')

    def red(self):
        self.canvas.itemconfigure(self.light, fill='red')

    async def _flash_callback(self):
        if self._flash:
            self._flash = False
            self.canvas.itemconfigure(self.light, fill=self._fill)
        else:
            self._flash = True
            self.canvas.itemconfigure(self.light, fill='darkgray')

    async def flash(self, fill: str = 'green', start: bool = None, stop: bool = None):
        '''Start or stop flashing'''
        if start and not self._timer:
            self._fill = fill
            self._timer = Timer(0.5, self._flash_callback, repeat=True)
        if stop and self._timer:
            self._timer.cancel()


class Timer:
    '''Async Timer with optional repeat'''

    def __init__(self, timeout, callback, repeat: bool = False):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.ensure_future(self._job())
        self._repeat = repeat

    async def _job(self):
        await asyncio.sleep(self._timeout)
        await self._callback()
        if self._repeat:
            self._task = asyncio.ensure_future(self._job())

    def cancel(self):
        '''Cancel task'''
        self._task.cancel()



'''Assetto Corsa server.cfg helper Class'''

import configparser
import logging
import os
import shutil


from ac_websocket_server.objects import SessionEvent
from ac_websocket_server.error import WebsocketsServerError


class ServerConfiguration:
    '''The server configuration'''

    def __init__(self, file_name: str = None) -> None:
        '''
        Create a new ServerConfiguration.
        '''

        self._logger = logging.getLogger('ac-ws.configuration')

        self._cfg = configparser.ConfigParser()
        self._cfg.optionxform = str

        self._dirty = False

        self._file_name = file_name
        self._dir_name = os.path.dirname(self._file_name)

        if not os.path.exists(self._file_name):
            error_message = f'Missing server_cfg.ini file in {self._dir_name}'
            self._logger.error(error_message)
            raise WebsocketsServerError(error_message)

        try:
            self._cfg.read(self._file_name)
            if self._cfg.has_section('QUALIFICATION'):
                self.rename_section(self._cfg, 'QUALIFICATION', 'QUALIFY')
            self._cfg['QUALIFY']['NAME'] = 'Qualify'
        except configparser.Error as e:
            error_message = f'Unable to parse server_cfg.ini file in {self._dir_name}'
            raise WebsocketsServerError(error_message) from e

    @property
    def cars(self):
        '''Allowed cars'''
        return self._cfg['SERVER']['CARS']

    @cars.setter
    def cars(self, value):
        self._cfg['SERVER']['CARS'] = value
        self._dirty = True

    @property
    def http_port(self):
        '''HTTP port'''
        return self._cfg['SERVER']['HTTP_PORT']

    @property
    def name(self):
        '''Name of server'''
        return self._cfg['SERVER']['NAME']

    @name.setter
    def name(self, value):
        self._cfg['SERVER']['NAME'] = value
        self._dirty = True

    def rename_section(self, cfg, section_from, section_to):
        '''https://stackoverflow.com/questions/15069127/python-configparser-module-rename-a-section'''
        items = cfg.items(section_from)
        cfg.add_section(section_to)
        for item in items:
            cfg.set(section_to, item[0], item[1])
        cfg.remove_section(section_from)

    def session_disable(self, session_name: str):
        '''Disable a session'''

        session_name = session_name.upper()

        if self._cfg.has_section(session_name):
            self._cfg.remove_section(session_name)
            self._dirty = True

    def session_enable(self, session_name: str):
        '''Enable a session'''

        session_name = session_name.upper()

        if not self._cfg.has_section(session_name):
            self._cfg.add_section(session_name)
            self._cfg.set(session_name, 'NAME', session_name.capitalize())
            self._cfg.set(session_name, 'IS_OPEN', '1')
            if session_name == 'PRACTICE' or session_name == 'QUALIFY':
                self._cfg.set(session_name, 'TIME', '10')
            if session_name == 'RACE':
                self._cfg.set(session_name, 'LAPS', '10')
                self._cfg.set(session_name, 'WAIT_TIME', '60')
            self._dirty = True

    def session_modify(self, session_name: str,
                       laps: int | None = None,
                       time: int | None = None):
        '''Modify a session'''

        session_name = session_name.upper()

        if laps and session_name == 'RACE':
            self._cfg.set(session_name, 'LAPS', laps)
            self._cfg.remove_option(session_name, 'TIME')
            self._dirty = True
        if time:
            self._cfg.set(session_name, 'TIME', time)
            self._cfg.remove_option(session_name, 'LAPS')
            self._dirty = True

    @property
    def sessions(self):
        '''Dict of sessions'''

        sessions = {}

        for session in ['PRACTICE', 'QUALIFY', 'RACE']:
            if self._cfg.has_section(session):
                name = self._cfg[session].get('NAME').capitalize()
                time = self._cfg[session].get('TIME', 0)
                laps = self._cfg[session].get('LAPS', 0)
                sessions[name] = SessionEvent(name, laps=laps, time=time)

        return sessions

    @property
    def tcp_port(self):
        '''TCP port'''
        return self._cfg['SERVER']['TCP_PORT']

    @property
    def track(self):
        '''Name of track'''
        return self._cfg['SERVER']['TRACK']

    @track.setter
    def track(self, value):
        self._cfg['SERVER']['TRACK'] = value
        self._dirty = True

    @property
    def udp_port(self):
        '''UDP port'''
        return self._cfg['SERVER']['UDP_PORT']

    def write(self):
        '''Write the server config file'''
        # pylint: disable=logging-fstring-interpolation

        if not self._dirty:
            self._logger.error(f'{self._file_name} is not dirty')

        try:

            if not os.path.exists(self._file_name + '.old'):
                shutil.copy(self._file_name, self._file_name + '.old')
                self._logger.debug(
                    f'Created {self._file_name}.old before changes')

            with open(self._file_name, 'w', encoding='utf-8') as f:
                self._cfg.write(f, space_around_delimiters=False)
                self._dirty = False

        except (IOError, OSError) as e:
            error_message = f'Unable to write server_cfg.ini file in {self._dir_name}'
            self._logger.error(error_message)
            raise WebsocketsServerError(error_message) from e



'''Assetto Corsa Websocket Connection Class'''

import asyncio
import json
import logging
import os
from packaging.version import Version
from typing import Any
import websockets

from ac_websocket_server.constants import PROTOCOL
from ac_websocket_server.objects import LogonInfo
from ac_websocket_server.observer import Observer, ObserverNotifier
from ac_websocket_server.protocol import Protocol

EXTRA_DEBUG = False


class WebSocketConnection(ObserverNotifier):
    '''Represents an individual webscocket connection'''

    def __init__(self, server: Any = None, websocket: Any = None) -> None:

        super().__init__()

        self._logger = logging.getLogger('ac-ws.ws-connection')

        self._protocol_version = None

        self._server = server

        if self._server.game:
            self._server.game.subscribe(self)
            self._server.game.grid.subscribe(self)
        if self._server.tracker:
            self._server.tracker.subscribe(self)

        self.websocket = websocket

    def close(self):
        '''Close and cleanup subscriptions'''
        if self._server.game:
            self._server.game.unsubscribe(self)
            self._server.game.grid.unsubscribe(self)
        if self._server.tracker:
            self._server.tracker.unsubscribe(self)

    async def consumer(self, message):
        '''ACWS consumer function for all received messages from client'''
        # pylint: disable=logging-fstring-interpolation, line-too-long

        if message == b'\n':
            return

        self._logger.debug(f'Received message: {message}')

        if isinstance(message, bytes):
            message_string = str(message.strip(), 'utf-8')
        else:
            message_string = message.strip()

        message_words = message_string.split()

        if message_words[0] == 'logon:':
            login_info = json.loads(''.join(message_words[1:]))
            self._protocol_version = login_info['protocol']
            self._logger.debug(
                f'Logon from client with protocol version {self._protocol_version}')
            return

        if not self._protocol_version or Version(self._protocol_version) < Version(PROTOCOL):
            msg = f'Command ignored - client not compatible with PROTOCOL version: {PROTOCOL}'
            await self.send(Protocol.error(msg=msg))
            return

        message_funcs = {'grid': self._server.game.grid.consumer,
                         'lobby': self._server.game.lobby.consumer,
                         'server': self._server.game.consumer,
                         'shutdown': self._server.shutdown,
                         'tracker': self._server.tracker.consumer}

        if message_funcs.get(message_words[0]) and len(message_words) > 1:
            await message_funcs[message_words[0]](message_words[1:], connection=self)
            return

        await self.send(Protocol.error(
            msg=f'Received unrecognised message: {message}'))

    async def producer(self):
        '''Pull a message off the queue and send on websocket'''
        # pylint: disable=logging-fstring-interpolation

        data = await self._notifier_queue.get()
        self._logger.debug(f'Sending message: {data}')
        if data:
            return data
        return ''

    async def send(self, msg: str):
        '''Send a message on the websocket by pushing it to notify queue'''

        await self._notifier_queue.put(msg)



#!/usr/bin/env python

'''Console UI'''

import textwrap
import tkinter as tk

from ac_websocket_client.objects import (
    CONSOLE_X, GriddedButton, GriddedEntry, GriddedFrame,
    GriddedLabel, GriddedListbox, GriddedText, TrafficLight)


class ConsoleUI(GriddedFrame):
    '''Console UI'''

    ENABLE_WORDWRAP = False

    def __init__(self, parent):

        super().__init__(grid_row=4, grid_col=0, height_by=3)

        self.parent = parent

        self.configure_columns(1, 1, 4, 1, 1)

        GriddedLabel(self, grid_row=0, grid_col=0, width=8, text="Console")
        GriddedLabel(self, grid_row=0, grid_col=1, width=8, text="command:")

        self._field = tk.StringVar()
        self.entry = GriddedEntry(self, grid_row=0, grid_col=2,
                                  textvariable=self._field)
        self.entry.bind('<Return>', self._async_send_command)

        self._button = tk.StringVar(value='Send')
        GriddedButton(self, grid_row=0, grid_col=3,
                      textvariable=self._button,
                      command=lambda: self.parent.loop.create_task(self.parent.send_command()))

        self._light = TrafficLight(self, row=0, column=4)
        self._light.gray()

        self._listbox = GriddedListbox(
            self, grid_row=1, grid_col=0, grid_span=5)

        self.update_ui()

    def _async_send_command(self, _event):
        '''Send command to ACWS server'''
        self.parent.loop.create_task(self.parent.send_command())

    def input(self):
        '''Return command input to be sent to ACWS server'''
        return self._field.get()

    def output(self, timestamp: str, lines: str, fg='Black'):
        '''Send command output to console.'''

        indent = '.' * len(timestamp)
        output_fmt = {'fg': fg}

        for input_line in lines.splitlines():
            output_lines = f'{timestamp}: {input_line}\n'
            if ConsoleUI.ENABLE_WORDWRAP:
                for line in textwrap.wrap(output_lines,
                                          width=CONSOLE_X,
                                          subsequent_indent=indent):
                    self._listbox.insert(tk.END, line)
                    self._listbox.itemconfig(tk.END, output_fmt)
            else:
                self._listbox.insert(tk.END, output_lines)
                self._listbox.itemconfig(tk.END, output_fmt)

        self.update_ui()

    def update_ui(self):
        '''Update the UI'''

        self._listbox.yview(tk.END)



#!/usr/bin/env python

'''Server UI'''

import re
import tkinter as tk
from typing import Dict

from ac_websocket_client.objects import (
    GriddedButton, GriddedEntry, GriddedFrame, GriddedLabel, TrafficLight)


class ServerUI(GriddedFrame):
    '''Server UI'''

    def __init__(self, parent):

        super().__init__(grid_row=1, grid_col=0, height_by=3.5)

        self.parent = parent

        self.configure_columns(1, 1, 4, 1, 1)

        self._buttons = {}
        self._fields = {}
        self._lights = {}

        grid_row = 0

        GriddedLabel(self, grid_row=grid_row, grid_col=0,
                     width=8, text="Game")

        GriddedLabel(self, grid_row=grid_row, grid_col=1,
                     width=8, text="started:")
        self._fields['started'] = tk.StringVar()
        GriddedEntry(self, grid_row=grid_row, grid_col=2,
                     textvariable=self._fields['started'], state=tk.DISABLED)
        self._buttons['game'] = tk.StringVar(value='Start Game')
        GriddedButton(self, grid_row=grid_row, grid_col=3,
                      textvariable=self._buttons['game'],
                      command=lambda: self.parent.loop.create_task(self.parent.toggle_game()))
        self._lights['game'] = TrafficLight(self, row=grid_row, column=4)

        grid_row += 1

        self._fields['registered'] = tk.StringVar()
        GriddedLabel(self, grid_row=grid_row, grid_col=1,
                     width=8, text="registered:")
        GriddedEntry(self, grid_row=grid_row, grid_col=2,
                     textvariable=self._fields['registered'], state=tk.DISABLED)
        self._lights['registered'] = TrafficLight(self, row=grid_row, column=4)
        self._buttons['lobby'] = tk.StringVar(value='(Re)register')
        GriddedButton(self, grid_row=grid_row, grid_col=3,
                      textvariable=self._buttons['lobby'],
                      command=lambda: self.parent.loop.create_task(self.parent.toggle_registration()))
        self._lights['lobby'] = TrafficLight(self, row=grid_row, column=4)

        grid_row += 1

        GriddedLabel(self, grid_row=grid_row,
                     grid_col=0, width=8, text="Config")

        self._fields['cfg'] = tk.StringVar()
        GriddedLabel(self, grid_row=grid_row, grid_col=1, width=8, text="cfg:")
        GriddedEntry(self, grid_row=grid_row, grid_col=2,
                     textvariable=self._fields['cfg'], state=tk.DISABLED)
        self._buttons['save'] = tk.StringVar(value='Save .cfg')
        GriddedButton(self, grid_row=grid_row, grid_col=3,
                      textvariable=self._buttons['save'],
                      command=lambda: self.parent.loop.create_task(self.parent.save_configuration()))
        self._lights['save'] = TrafficLight(self, row=grid_row, column=4)

        grid_row += 1

        self._fields['name'] = tk.StringVar()
        GriddedLabel(self, grid_row=grid_row,
                     grid_col=1, width=8, text="name:")
        GriddedEntry(self, grid_row=grid_row, grid_col=2,
                     textvariable=self._fields['name'], state=tk.DISABLED)

        grid_row += 1

        self._fields['track'] = tk.StringVar()
        GriddedLabel(self, grid_row=grid_row,
                     grid_col=1, width=8, text="track:")
        GriddedEntry(self, grid_row=grid_row, grid_col=2,
                     textvariable=self._fields['track'], state=tk.DISABLED)

        grid_row += 1

        self._fields['cars'] = tk.StringVar()
        GriddedLabel(self, grid_row=grid_row,
                     grid_col=1, width=8, text="cars:")
        GriddedEntry(self, grid_row=grid_row, grid_col=2,
                     textvariable=self._fields['cars'], state=tk.DISABLED)

        self._buttons['sessions'] = {}
        self._fields['sessions'] = {}
        self._lights['sessions'] = {}

        grid_row += 1

        self._fields['sessions']['Practice'] = tk.StringVar()
        self._fields['sessions']['Practice'].trace_add(
            'write', self._trace_sessions)
        GriddedLabel(self, grid_row=grid_row, grid_col=1,
                     width=8, text="practice:")
        GriddedEntry(self, grid_row=grid_row, grid_col=2,
                     textvariable=self._fields['sessions']['Practice'])
        self._buttons['sessions']['Practice'] = tk.StringVar(value='Enable')
        GriddedButton(self, grid_row=grid_row, grid_col=3,
                      textvariable=self._buttons['sessions']['Practice'],
                      command=lambda: self.parent.loop.create_task(self._handle_session_button('Practice')))
        self._lights['sessions']['Practice'] = TrafficLight(
            self, row=grid_row, column=4)

        grid_row += 1

        self._fields['sessions']['Qualify'] = tk.StringVar()
        self._fields['sessions']['Qualify'].trace_add(
            'write', self._trace_sessions)
        GriddedLabel(self, grid_row=grid_row, grid_col=1,
                     width=8, text="qualify:")
        GriddedEntry(self, grid_row=grid_row, grid_col=2,
                     textvariable=self._fields['sessions']['Qualify'])
        self._buttons['sessions']['Qualify'] = tk.StringVar(value='Enable')
        GriddedButton(self, grid_row=grid_row, grid_col=3,
                      textvariable=self._buttons['sessions']['Qualify'],
                      command=lambda: self.parent.loop.create_task(self._handle_session_button('Qualify')))
        self._lights['sessions']['Qualify'] = TrafficLight(
            self, row=grid_row, column=4)

        grid_row += 1

        self._fields['sessions']['Race'] = tk.StringVar()
        self._fields['sessions']['Race'].trace_add(
            'write', self._trace_sessions)
        GriddedLabel(self, grid_row=grid_row, grid_col=1,
                     width=8, text="race:")
        GriddedEntry(self, grid_row=grid_row, grid_col=2,
                     textvariable=self._fields['sessions']['Race'])
        self._buttons['sessions']['Race'] = tk.StringVar(value='Enable')
        GriddedButton(self, grid_row=grid_row, grid_col=3,
                      textvariable=self._buttons['sessions']['Race'],
                      command=lambda: self.parent.loop.create_task(self._handle_session_button('Race')))
        self._lights['sessions']['Race'] = TrafficLight(
            self, row=grid_row, column=4)

        self.update_ui()

    async def _handle_session_button(self, session: str):
        '''Handle session button presses for both 'Apply' and Enable/Disable'''

        if self._buttons['sessions'][session].get() == 'Apply':

            session_details = self._parse_session_description(
                self._fields['sessions'][session].get())
            if session_details.get('laps', None):
                session_type = 'laps'
                session_value = session_details['laps']
            if session_details.get('time', None):
                session_type = 'time'
                session_value = session_details['time']

            await self.parent.update_session(session,
                                             session_type,
                                             session_value)

        else:

            await self.parent.toggle_session(session)

    def _parse_session_description(self, description: str) -> Dict[str, int]:
        '''Parse the session description and return a dictionary '''

        result = {}

        m = re.compile(r'^(\d*) (\S*)$').match(description)
        if m:
            session_value = m.group(1)
            session_type = m.group(2)
        if session_type == 'laps':
            result['laps'] = int(session_value)
        if session_type == 'minutes':
            result['time'] = int(session_value)

        return result

    def _trace_sessions(self, *args):
        '''Update the parent with updated values'''

        for session in ('Practice', 'Qualify', 'Race'):

            if old_session := self.parent.sessions.get(session, None):
                old_laps = int(old_session['laps'])
                old_time = int(old_session['time'])
            else:
                old_laps = None
                old_time = None

            m = re.compile(
                r'^(\d*) (\S*)$').match(self._fields['sessions'][session].get())
            if m:
                new_value = m.group(1)
                if new_value != '':
                    new_value = int(new_value)
                new_type = m.group(2)
                if new_type == 'laps' and old_laps != new_value:
                    self._buttons['sessions'][session].set('Apply')
                if new_type == 'minutes' and old_time != new_value:
                    self._buttons['sessions'][session].set('Apply')

    def update_ui(self):
        '''Update the UI with the contents of the parent.server'''

        self._fields['cfg'].set(self.parent.server.get('child_ini_file', ''))
        self._fields['name'].set(self.parent.server.get('name', ''))
        self._fields['track'].set(self.parent.server.get('track', ''))
        self._fields['cars'].set(self.parent.server.get('cars', ''))

        if self.parent.states.is_connected:

            if self.parent.states.is_started:
                self._fields['started'].set(
                    self.parent.server.get('timestamp', None))
                self._buttons['game'].set('Stop Game')
                self._lights['game'].green()
            else:
                self._fields['started'].set('n/a')
                self._buttons['game'].set('Start Game')
                self._lights['game'].red()

            if self.parent.states.is_registered:
                self._fields['registered'].set(
                    self.parent.lobby.get('since', ''))
                self._buttons['lobby'].set('Re-register')
                self._lights['lobby'].green()
            else:
                self._fields['registered'].set('n/a')
                self._buttons['lobby'].set('Register')
                self._lights['lobby'].red()

        else:

            self._lights['game'].gray()
            self._lights['lobby'].gray()

        if self.parent.states.cfg_needs_saving:
            self._lights['save'].red()
        else:
            self._lights['save'].gray()

        for session in ('Practice', 'Qualify', 'Race'):
            self._buttons['sessions'][session].set('Enable')
            self._fields['sessions'][session].set('')
            self._lights['sessions'][session].gray()

        if sessions := self.parent.sessions:
            for session_type in sessions:
                if session_type in ('Practice', 'Qualify', 'Race'):
                    self._buttons['sessions'][session_type].set('Disable')
                    session_active = sessions[session_type]['active']
                    if session_active:
                        self._lights['sessions'][session_type].green()
                    else:
                        self._lights['sessions'][session_type].gray()
                    if sessions[session_type]['laps'] == 0:
                        session_description = f'{str(sessions[session_type]["time"])} minutes'
                    else:
                        session_description = f'{str(sessions[session_type]["laps"])} laps'
                    self._fields['sessions'][session_type].set(
                        session_description)

        self.parent.loop.create_task(self._update_ui_async())

    async def _update_ui_async(self):
        '''Update async parts of UI'''
        if self.parent.states.cfg_needs_reload:
            self._fields['cfg'].set(self.parent.server.get(
                'child_ini_file', '') + '- needs server reload')
            await self._lights['save'].flash(start=True)
        else:
            await self._lights['save'].flash(stop=True)



'''Assetto Corsa Grid Class'''

from datetime import datetime
import glob
import logging
import os
import shutil
from typing import Any, List

from ac_websocket_server.debug import DebugTransaction
from ac_websocket_server.entries import EntryList
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.observer import Notifier
from ac_websocket_server.protocol import Protocol

ENTRY_LIST_INI = 'entry_list.ini'
ENTRY_LIST_NEW = 'entry_list_new.ini'
ENTRY_LIST_OLD = 'entry_list_old.ini'


class Grid(Notifier):
    '''Represents grid setup'''

    def __init__(self, server_directory: str, track: str = None, entry_list: EntryList = None):

        self._logger = logging.getLogger('ac-ws.grid')

        self.server_directory = server_directory

        self._debug_transaction = DebugTransaction(self.server_directory)

        self.entry_list = entry_list
        self.track = track

        super().__init__()

    async def consumer(self, message_words: List[str], connection: id = None):
        '''Consume args destined for the server'''

        message_funcs = {'finish': self.__update_finishing,
                         'reverse': self.__update_reversed,
                         'entries': self.__entries,
                         'order': self.__order,
                         'save': self.__save}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]](connection=connection)

    async def __entries(self, connection: Any = None):
        '''Show original grid order info as a JSON string'''
        await self.put(Protocol.success({'entries': self.entry_list.show_entries()}), observer=connection)

    def __debug_transaction(self, name: str):
        '''Save all grid files in debug transaction'''

        self._debug_transaction.open(name)
        self._debug_transaction.save_file(
            f'{self.server_directory}/cfg/{ENTRY_LIST_INI}')
        self._debug_transaction.save_file(
            f'{self.server_directory}/cfg/{ENTRY_LIST_NEW}')
        self._debug_transaction.save_file(
            f'{self.server_directory}/cfg/{ENTRY_LIST_OLD}')
        self._debug_transaction.close()

    async def __order(self, connection: Any = None):
        '''Show current grid order info as a JSON string'''
        await self.put(Protocol.success({'grid': self.entry_list.show_grid()}), observer=connection)

    def restore(self, connection: Any = None):
        '''Restore updated entry_list.ini file'''

        self.__debug_transaction('grid-restore-pre')

        try:
            if os.path.exists(f'{self.server_directory}/cfg/{ENTRY_LIST_NEW}'):
                os.replace(f'{self.server_directory}/cfg/{ENTRY_LIST_INI}',
                           f'{self.server_directory}/cfg/{ENTRY_LIST_OLD}')
                os.replace(f'{self.server_directory}/cfg/{ENTRY_LIST_NEW}',
                           f'{self.server_directory}/cfg/{ENTRY_LIST_INI}')
        except OSError as error:
            self._logger.error(f'Restore failed: {error}')

        self.__debug_transaction('grid-restore-post')

    async def __save(self, connection: Any = None):
        '''Write updated entry_list.ini file'''

        self.__debug_transaction('grid-save-pre')

        try:
            self.entry_list.write(
                f'{self.server_directory}/cfg/{ENTRY_LIST_NEW}')
            await self.put(Protocol.success(msg='entry_list_new.ini file update SUCCESS'), observer=connection)
        except WebsocketsServerError as error:
            await self.put(Protocol.error(
                msg=f'entry_list.ini file update FAILURE: {error}'), observer=connection)

        self.__debug_transaction('grid-save-post')

    async def update(self, reverse: bool = False, connection: Any = None):
        '''Update grid based on latest results.JSON file and using sorting rule'''

        if reverse:
            self._debug_transaction.open('grid-reverse')
            self.entry_list.set_reversed_order()
        else:
            self._debug_transaction.open('grid-finish')
            self.entry_list.set_standard_order()

        try:
            result_files = glob.glob(
                os.path.join(f'{self.server_directory}/results', "*RACE.json"))
            if len(result_files) > 0:
                result_file = max(result_files, key=os.path.getctime)
                self.entry_list.parse_result_file(
                    result_file, track=self.track)
                self._debug_transaction.save_file(
                    f'{self.server_directory}/cfg/entry_list.ini')
                self._debug_transaction.save_file(result_file)
            else:
                await self.put(Protocol.error(
                    msg='Unable to parse results file - no file exists'), observer=connection)
                return
        except WebsocketsServerError as error:
            await self.put(Protocol.error(
                msg=f'{result_file} parse FAILURE: {error}'), observer=connection)
            return

        await self.put(Protocol.success(
            msg=f'{result_file} parse SUCCESS'), observer=connection)

        await self.__order(connection=connection)

        self._debug_transaction.close()

    async def __update_finishing(self, connection: Any = None):
        '''Update grid in standard finishing order'''
        await self.update(connection=connection)

    async def __update_reversed(self, connection: Any = None):
        '''Update grid in reverse finishing order'''
        await self.update(reverse=True, connection=connection)



'''Assetto Corsa Lobby Class'''

import aiohttp
import asyncio
import json
import logging
import os
import sys
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List

from ac_websocket_server.objects import EnhancedJSONEncoder, MessageType, LobbyEvent
from ac_websocket_server.observer import Notifier
from ac_websocket_server.protocol import Protocol


@dataclass
class Lobby(Notifier):
    '''Represents an Assetto Corsa Lobby'''

    connected: bool = field(default=False, init=False)
    since: datetime = field(default=None, init=False)
    url_register: str = field(default=None, init=False)
    url_ping: str = field(default=None, init=False)

    def __post_init__(self):

        super().__init__()

        self._logger = logging.getLogger('ac-ws.lobby')

    async def consumer(self, message_words: List[str], connection: id = None):
        '''Consume args destined for the lobby'''

        message_funcs = {'info': self._info,
                         'restart': self._restart}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]]()

    async def _info(self):
        '''Send lobby information'''
        await self.put(Protocol.success({'lobby': self}))

    async def _restart(self):
        '''Re-start the lobby connection'''

        self._logger.info('Re-regestering to lobby')
        async with aiohttp.ClientSession() as session:

            if not self.url_register:
                msg = f'Cannot attempt to re-register with lobby as no URL is defined'
                self._logger.error(msg)
                await self.put(Protocol.error(msg=msg))
                return

            async with session.get(self.url_register) as resp:
                text = await resp.text()
                if resp.status == 200 and 'ERROR' not in text:
                    self._logger.info('Successfully re-registered to lobby')
                    self.connected = True
                    self.since = datetime.now().strftime("%Y%m%d_%H%M%S")
                    lobby_event = LobbyEvent(
                        connected=self.connected,
                        timestamp=self.since,
                        url_register=self.url_register,
                        url_ping=self.url_ping)
                    await self.put(json.dumps(Message(type=MessageType.LOBBY_EVENT,
                                                      body=lobby_event),
                                              cls=EnhancedJSONEncoder))
                    # await self.put(Protocol.success(msg=text))

                else:
                    self._logger.error(
                        f'Failed attempt to re-register with lobby: {text}')
                    await self.put(Protocol.error(msg=text))

    async def update(self, lobby_event: LobbyEvent):
        '''Update lobby info'''

        if lobby_event['connected'] and not self.connected:
            self.connected = True
            self.since = lobby_event.get(
                'timestamp', datetime.now().strftime("%Y%m%d_%H%M%S"))

        if not lobby_event['connected'] and self.connected:
            self.connected = False
            self.since = ''
            await self._restart()

        if lobby_event['url_register']:
            self.url_register = lobby_event['url_register']

        if lobby_event['url_ping']:
            self.url_ping = lobby_event['url_ping']
